/*
 * rbt.c
 *
 *  Created on: Apr 26, 2011
 *      Author: tony
 */

/*
    License

    This file is part of ipclassify.

    ipclassify is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    ipclassify is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ipclassify.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <config.h>



/* ----- Extern ------------------------------------------------------- */
extern bool debug;

/* -------------------------------------------------------------------- */

/*
 * return 0 for matching keys,
 * > 0 if a > b
 * < 0 if a < b
 *
 * a is the tree key
 * b is the key being looked for or inserted
 */
int
compKey(keyType *a, keyType *b) {
	unsigned long leafmin = 0, leafmax = 0;
	unsigned long keymin = 0, keymax = 0;
	/*
	 *  two conditions need to be checked:
	 *  1. if b is only one long in the lower bits.
	 *  2. b has an ipmax in the higher bits and ipmin in the lower.
	 */

	leafmin = getlow(a->key);
	leafmax = gethigh(a->key);
	keymin = getlow(b->key);
	keymax = gethigh(b->key);

	if(b->RangeCheck == TRUE) {
		if(keymax == 0 || keymax == 0xFFFFFF) {
			if(leafmax < keymin) { return(-1); }
			if(leafmin > keymin) { return(+1); }
			return(0);
		} else {
			if((leafmin > keymax)) { return(+1); }
			if((leafmax < keymin)) { return(-1); }
			return(0);
		}
	} else {
		if(a->key < b->key) { return(-1);}
		if(b->key < a->key) { return(+1);}
		return(0);
	}
}

keyType *
copyKey(keyType *key) {
  keyType *tmpkey = NULL;

  if((tmpkey = calloc(sizeof(keyType),1)) == NULL) {
    syslog(LOG_ERR,"calloc() falure in copyKey()");
    log2file("calloc() falure in copyKey()");
    debugMsg("calloc() falure in copyKey()",1);
    Terminate(EXIT_FAILURE);
  }
  tmpkey->key = key->key;
  tmpkey->RangeCheck = key->RangeCheck;
  return(tmpkey);
}

recType *
copyValue(recType *value) {
  recType *tmpRec = NULL;
  if((tmpRec = calloc(sizeof(recType),1)) == NULL) {
    syslog(LOG_ERR,"calloc() falure in copyValue()");
    log2file("calloc() falure in copyValue()");
    debugMsg("calloc() falure in copyValue()",1);
    Terminate(EXIT_FAILURE);
  }
  tmpRec->hits = value->hits;
  tmpRec->blockname = value->blockname;
  tmpRec->key = value->key;
  return(tmpRec);
}


void
destroyKey(keyType *key) {
  free(key);
}

void
destroyValue(recType *value) {
  free(value->blockname);
  free(value);
}



int
addToRBT(queue_t *node,char *name,char *ipmin,char *ipmax, int whitelist){
	char mstr[DEFBUFSIZE];
	recType rec ={NULL,0,0};
	recType *recptr = NULL;
	keyType key = {0,TRUE};
	unsigned long long int_ipmin = 0,int_ipmax = 0;
	cp_rbtree *whichRBT = NULL;

    rec.hits = 0;
    int_ipmax = ntohl(inet_addr(ipmax));
    int_ipmin = ntohl(inet_addr(ipmin));
    /* zero out the higher bits, ntohl() doesen't do that */
    int_ipmin = int_ipmin & LOWMASK;
    int_ipmax = int_ipmax & LOWMASK;
    key.key = sethigh(int_ipmax);
    key.key = setlow(key.key,int_ipmin);

    /* if whitelist is true then add to the whitelist vice the rbt */
    if(whitelist) {
    	whichRBT = node->whitelist;
    } else {
    	whichRBT = node->rbt;
    }

    /* need to initialize rbt if it has not been done yet */
	if(whichRBT == NULL){
		whichRBT = cp_rbtree_create_by_option(COLLECTION_MODE_COPY |
				COLLECTION_MODE_DEEP,
				(cp_compare_fn) compKey,
				(cp_copy_fn) copyKey,
				(cp_destructor_fn) destroyKey,
				(cp_copy_fn) copyValue,
				(cp_destructor_fn) destroyValue);
		if(whichRBT == NULL) {
			/* If there is a failure to create the tree, Bye-Bye */
		    syslog(LOG_ERR,"cp_rbtree_creaate() falure in addToRBT()");
		    log2file("cp_rbtree_creaate() falure in addToRBT()");
		    debugMsg("cp_rbtree_creaate() falure in addToRBT()",1);
		    Terminate(EXIT_FAILURE);
		}
		if((rec.blockname = calloc(strlen(name)+1,1)) == NULL) {
		    syslog(LOG_ERR,"calloc() falure in addToRBT() while stringing");
		    log2file("calloc() falure in addToRBT() while stringing");
		    debugMsg("calloc() falure in addToRBT() while stringing",1);
		    Terminate(EXIT_FAILURE);
		}
		strncpy(rec.blockname,name,strlen(name));
		rec.key = key.key;
	     if(cp_rbtree_insert(whichRBT,&key,&rec) == NULL) {
		        syslog(LOG_ERR,"cp_rbtree_insert() failure in addToRBT()");
		        log2file("cp_rbtree_insert() failure in addToRBT()");
		        debugMsg("cp_rbtree_insert() failure in addToRBT()",1);
		        Terminate(EXIT_FAILURE);
	     }
	     /*
	      * once inserted we need to free the nolonger needed rec because it
	      * was copied into the rbt.
	      */
	     if(whitelist) { node->whitelist = whichRBT; }
	     else { node->rbt = whichRBT; }
		 snprintf((char *)mstr,DEFBUFSIZE, "First Node Added \"%s\"",
				rec.blockname);
			/* OK, this is redundant, but saves a function call overhead when not in foreground */
		 if(debug) debugMsg(mstr,2);
	     return(SUCCESS);
	} else {
		/*
		 * Running cp_rbtree_get(key) will return non-null if our key is included
		 * within the ipmin/ipmax range.
		 */
		if((recptr = cp_rbtree_get(whichRBT,&key)) == NULL) {
			/* No merging to be done here, need to add new*/
			if((rec.blockname = calloc(strlen(name)+1,1)) == NULL) {
			    syslog(LOG_ERR,"calloc() falure in addToRBT() while stringing");
			    log2file("calloc() falure in addToRBT() while stringing");
			    debugMsg("calloc() falure in addToRBT() while stringing",1);
			    Terminate(EXIT_FAILURE);
			}
			strncpy(rec.blockname,name,strlen(name));
			rec.key = key.key;
		    if(cp_rbtree_insert(whichRBT,&key,&rec) == NULL) {
			        syslog(LOG_ERR,"cp_rbtree_insert() failure in addToRBT()");
			        log2file("cp_rbtree_insert() failure in addToRBT()");
			        debugMsg("cp_rbtree_insert() failure in addToRBT()",1);
			        Terminate(EXIT_FAILURE);
		    }
		     /*
		      * once inserted we need to free the nolonger needed rec because it
		      * was copied into the rbt.
		      */
			snprintf((char *)mstr,DEFBUFSIZE, "New Node Added \"%s\"",
					rec.blockname);
			/* OK, this is redundant, but saves a function call overhead when not in foreground */
			if(debug) debugMsg(mstr,2);
		    return(SUCCESS);
		} else {
			/*
			 * recptr should point to an existing node that we have to merge with.
			 * We do have to unlock the rbt tree, becaue cp_rbtree_get will lock.
			 *
			 * */
			unsigned long long leaf_ipmin = 0,leaf_ipmax = 0;
			leaf_ipmax = gethigh(recptr->key);
			leaf_ipmin = getlow(recptr->key);
			if((leaf_ipmin <= int_ipmin) && (int_ipmin <= leaf_ipmax)) {
				if(int_ipmax > leaf_ipmax) {
				    key.key = sethigh(int_ipmax);
				    key.key = setlow(key.key,leaf_ipmin);
					if((rec.blockname = calloc(strlen(recptr->blockname)+1,1)) == NULL) {
					    syslog(LOG_ERR,"calloc() failure in addToRBT() while stringing");
					    log2file("calloc() falure in addToRBT() while stringing");
					    debugMsg("calloc() falure in addToRBT() while stringing",1);
					    Terminate(EXIT_FAILURE);
					}
					strncpy(rec.blockname,recptr->blockname,strlen(recptr->blockname));
					rec.key = key.key;
					key.key = recptr->key;
					key.RangeCheck = FALSE;
					if(cp_rbtree_unlock(whichRBT) < 0){
				        syslog(LOG_ERR,"cp_rbtree_unlock() failure in addToRBT()");
				        log2file("cp_rbtree_unlock() failure in addToRBT()");
				        debugMsg("cp_rbtree_unlock() failure in addToRBT()",1);
				        Terminate(EXIT_FAILURE);
					}
					if(cp_rbtree_delete(whichRBT,&key) == NULL) {
				        syslog(LOG_ERR,"cp_rbtree_delete() failure in addToRBT()");
				        log2file("cp_rbtree_delete() failure in addToRBT()");
				        debugMsg("cp_rbtree_delete() failure in addToRBT()",1);
				        Terminate(EXIT_FAILURE);
					}
					key.key=rec.key;
				    if(cp_rbtree_insert(whichRBT,&key,&rec) == NULL) {
					        syslog(LOG_ERR,"cp_rbtree_insert() failure in addToRBT()");
					        log2file("cp_rbtree_insert() failure in addToRBT()");
					        debugMsg("cp_rbtree_insert() failure in addToRBT()",1);
					        Terminate(EXIT_FAILURE);
				    }
				    return(SUCCESS);
				} else {
					/* Skip Included Range */
					snprintf((char *)mstr,DEFBUFSIZE, "Skip Included Range \"%s\" inside \"%s\"",
							rec.blockname,recptr->blockname);
					debugMsg(mstr,2);
					if(cp_rbtree_unlock(whichRBT) < 0){
				        syslog(LOG_ERR,"cp_rbtree_unlock() failure in addToRBT()");
				        log2file("cp_rbtree_unlock() failure in addToRBT()");
				        debugMsg("cp_rbtree_unlock() failure in addToRBT()",1);
				        Terminate(EXIT_FAILURE);
					}
				    return(SUCCESS);
				}
			} else if ((leaf_ipmin <= int_ipmax) && (int_ipmax <= leaf_ipmax)) {
				if(leaf_ipmin > int_ipmin) {
				    key.key = sethigh(leaf_ipmax);
				    key.key = setlow(key.key,int_ipmin);
					if((rec.blockname = calloc(strlen(name)+1,1)) == NULL) {
					    syslog(LOG_ERR,"calloc() failure in addToRBT() while stringing");
					    log2file("calloc() falure in addToRBT() while stringing");
					    debugMsg("calloc() falure in addToRBT() while stringing",1);
					    Terminate(EXIT_FAILURE);
					}
					strncpy(rec.blockname,name,strlen(name));
					rec.key = key.key;
					key.key=recptr->key;
					key.RangeCheck = FALSE;
					if(cp_rbtree_unlock(whichRBT) < 0){
				        syslog(LOG_ERR,"cp_rbtree_unlock() failure in addToRBT()");
				        log2file("cp_rbtree_unlock() failure in addToRBT()");
				        debugMsg("cp_rbtree_unlock() failure in addToRBT()",1);
				        Terminate(EXIT_FAILURE);
					}
					if(cp_rbtree_delete(whichRBT,&key) == NULL) {
				        syslog(LOG_ERR,"cp_rbtree_delete() failure in addToRBT()");
				        log2file("cp_rbtree_delete() failure in addToRBT()");
				        debugMsg("cp_rbtree_delete() failure in addToRBT()",1);
				        Terminate(EXIT_FAILURE);
					}
					key.key = rec.key;
					key.RangeCheck = TRUE;
				    if(cp_rbtree_insert(whichRBT,&key,&rec) == NULL) {
					        syslog(LOG_ERR,"cp_rbtree_insert() failure in addToRBT()");
					        log2file("cp_rbtree_insert() failure in addToRBT()");
					        debugMsg("cp_rbtree_insert() failure in addToRBT()",1);
					        Terminate(EXIT_FAILURE);
				    }
				    return(SUCCESS);
				} else {
					/*  Skip Included Range */
					snprintf((char *)mstr,DEFBUFSIZE, "Skip Included Range \"%s\" with \"%s\"",
							rec.blockname,recptr->blockname);
					debugMsg(mstr,2);
					if(cp_rbtree_unlock(whichRBT) < 0){
				        syslog(LOG_ERR,"cp_rbtree_unlock() failure in addToRBT()");
				        log2file("cp_rbtree_unlock() failure in addToRBT()");
				        debugMsg("cp_rbtree_unlock() failure in addToRBT()",1);
				        Terminate(EXIT_FAILURE);
					}
				    return(SUCCESS);
				}
			} else {
				/* Replace the RBT node with the new one */
				if((rec.blockname = calloc(strlen(name)+1,1)) == NULL) {
				    syslog(LOG_ERR,"calloc() falure in addToRBT() while stringing");
				    log2file("calloc() falure in addToRBT() while stringing");
				    debugMsg("calloc() falure in addToRBT() while stringing",1);
				    Terminate(EXIT_FAILURE);
				}
				strncpy(rec.blockname,name,strlen(name));
				rec.key = key.key;
				key.key=recptr->key;
				key.RangeCheck = FALSE;
				if(cp_rbtree_unlock(whichRBT) < 0){
			        syslog(LOG_ERR,"cp_rbtree_unlock() failure in addToRBT()");
			        log2file("cp_rbtree_unlock() failure in addToRBT()");
			        debugMsg("cp_rbtree_unlock() failure in addToRBT()",1);
			        Terminate(EXIT_FAILURE);
				}
				if(cp_rbtree_delete(whichRBT,&key) == NULL) {
			        syslog(LOG_ERR,"cp_rbtree_delete() failure in addToRBT()");
			        log2file("cp_rbtree_delete() failure in addToRBT()");
			        debugMsg("cp_rbtree_delete() failure in addToRBT()",1);
			        Terminate(EXIT_FAILURE);
				}
				key.key = rec.key;
				key.RangeCheck = TRUE;
			    if(cp_rbtree_insert(whichRBT,&key,&rec) == NULL) {
				        syslog(LOG_ERR,"cp_rbtree_insert() failure in addToRBT()");
				        log2file("cp_rbtree_insert() failure in addToRBT()");
				        debugMsg("cp_rbtree_insert() failure in addToRBT()",1);
				        Terminate(EXIT_FAILURE);
			    }
			     /*
			      * once inserted we need to free the nolonger needed rec because it
			      * was copied into the rbt.
			      */
				snprintf((char *)mstr,DEFBUFSIZE, "Old Range replaced with \"%s\"",
						rec.blockname);
				if(debug) debugMsg(mstr,2);
			    return(SUCCESS);
			}
		}
	}

	return(SUCCESS);
}


int
findInRBT(queue_t *node, unsigned long long ip, recType *inrec) {
	recType *recptr = NULL;
	keyType key = {0,TRUE};
	int last_state = 0;

	ip = ip & LOWMASK;
	key.key = ip;

	if(node->rbt == NULL) return(FAILURE);

	/* cancellation is not allowed */
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,&last_state);
	if((node->whitelist != NULL) && (recptr = cp_rbtree_get(node->whitelist,&key)) != NULL) {
		/* in the whitelist */
		if(cp_rbtree_unlock(node->whitelist) < 0){
	        syslog(LOG_ERR,"cp_rbtree_unlock() failure in findInRBT()");
	        log2file("cp_rbtree_unlock() failure in findInRBT()");
	        debugMsg("cp_rbtree_unlock() failure in findInRBT()",1);
	        Terminate(EXIT_FAILURE);
		}
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,&last_state);
		pthread_testcancel();
		return(FAILURE);
	}
	if((node->rbt != NULL) && (recptr = cp_rbtree_get(node->rbt,&key)) == NULL) {
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,&last_state);
		pthread_testcancel();
		return(FAILURE);
	}
	/* if we are here we have a hit. */
	recptr->hits++;
	inrec->hits = recptr->hits;
	/* copy the block name */
	if((inrec->blockname = calloc(strlen(recptr->blockname)+1,1)) == NULL) {
	    syslog(LOG_ERR,"calloc() falure in findInRBT() while stringing");
	    log2file("calloc() falure in findInRBT() while stringing");
	    debugMsg("calloc() falure in findInRBT() while stringing",1);
	    Terminate(EXIT_FAILURE);
	}
	strncpy(inrec->blockname,recptr->blockname,strlen(recptr->blockname));
	if(cp_rbtree_unlock(node->rbt) < 0){
        syslog(LOG_ERR,"cp_rbtree_unlock() failure in findInRBT()");
        log2file("cp_rbtree_unlock() failure in findInRBT()");
        debugMsg("cp_rbtree_unlock() failure in findInRBT()",1);
        Terminate(EXIT_FAILURE);
	}
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,&last_state);
	pthread_testcancel();

	return(SUCCESS);
}


int RBTDumpStats(queue_t *head) {
	char mstr[DEFBUFSIZE];
	time_t t;
	struct tm *tm;
	char datestring[DEFBUFSIZE];
	queue_t *qptr = NULL;
	int tmpint = 0;

	t = time(NULL);
	tm = localtime(&t);
	if(tm == NULL) {
		snprintf((char *)mstr, DEFBUFSIZE, "RBTDumpStats(): localtime(): failure");
		syslog(LOG_ERR, mstr);
	    return(FAILURE);
	}
	strftime (datestring, sizeof(datestring), "%c", tm);

	log2stats("|---------------------------------------------------------------------------|");
	snprintf((char *)mstr, DEFBUFSIZE, "   Dumping Stats: %s",datestring);
	log2stats(mstr);
	log2stats("|---------------------------------------------------------------------------|");

	/* what if no queues ? */
	if ( head == NULL ) {
		log2stats("\tNothing to Dump: No Queues.");
		return(SUCCESS);
	}

	qptr = head;
	while(qptr != NULL) {
		log2stats("|---------------------------------------------------------------------------|");
		snprintf((char *)mstr, DEFBUFSIZE, "\t Queue: %d",qptr->queue);
		log2stats(mstr);

		if(qptr->rbt != NULL) {
			snprintf((char *)mstr, DEFBUFSIZE, "\t Total Ranges: %ld",cp_rbtree_count(qptr->rbt));
			log2stats(mstr);
		}
		if(qptr->whitelist != NULL) {
			snprintf((char *)mstr, DEFBUFSIZE, "\t Total Whitelist Ranges: %ld",cp_rbtree_count(qptr->whitelist));
			log2stats(mstr);
		}
		pthread_rwlock_wrlock(&qptr->rwlock);
	    tmpint = qptr->ttl_packets;
	    pthread_rwlock_unlock(&qptr->rwlock);
		snprintf((char *)mstr, DEFBUFSIZE, "\t Total Packets: %llu",tmpint);
		log2stats(mstr);
		pthread_rwlock_wrlock(&qptr->rwlock);
	    tmpint = qptr->ttl_rejected;
	    pthread_rwlock_unlock(&qptr->rwlock);
		snprintf((char *)mstr, DEFBUFSIZE, "\t Total Rejected: %llu",tmpint);
		log2stats(mstr);

		log2stats("\n\t - Hit Summary -\n");

		cp_rbtree_callback(qptr->rbt,statsDumpCallback,NULL);

		qptr = (queue_t *)qptr->next;
	}

	return(SUCCESS);
}



int statsDumpCallback(void *entry,void *prm){
	char mstr[DEFBUFSIZE];
	cp_rbnode *node = entry;
	recType *rec = node->value;

	if(rec->hits == 0) {return(0);}

	snprintf((char *)mstr, DEFBUFSIZE, "%4lu - %s",rec->hits,rec->blockname);
	log2stats(mstr);

	return(0);
}




int RBTResetStats(queue_t *head) {
	char mstr[DEFBUFSIZE];
	queue_t *qptr = NULL;

	/* what if no queues ? */
	if ( head == NULL ) {
		log2stats("\tNothing to Reset: No Queues.");
		return(SUCCESS);
	}

	qptr = head;
	while(qptr != NULL) {
		pthread_rwlock_wrlock(&qptr->rwlock);
	    qptr->ttl_packets = 0;
	    qptr->ttl_rejected = 0;
	    pthread_rwlock_unlock(&qptr->rwlock);
	    cp_rbtree_callback(qptr->rbt,statsResetCallback,NULL);
		qptr = (queue_t *)qptr->next;
	}
	return(SUCCESS);
}


int statsResetCallback(void *entry,void *prm){
	char mstr[DEFBUFSIZE];
	cp_rbnode *node = entry;
	recType *rec = node->value;

	if(rec->hits == 0) {return(0);}
	rec->hits = 0;
	return(0);
}
